package flink.zanxiangnet.ad.monitoring;

import flink.zanxiangnet.ad.monitoring.clickhouse.sink.BatchSinkHour;
import flink.zanxiangnet.ad.monitoring.clickhouse.sink.BatchSinkMinute;
import flink.zanxiangnet.ad.monitoring.pojo.entity.*;
import flink.zanxiangnet.ad.monitoring.process.*;
import flink.zanxiangnet.ad.monitoring.sink.TunnelBatchStreamSink;
import flink.zanxiangnet.ad.monitoring.stream.KeyedBatchStream;
import flink.zanxiangnet.ad.monitoring.trigger.AdMinuteODSStreamTrigger;
import flink.zanxiangnet.ad.monitoring.trigger.CostMinuteDMStreamTrigger;
import flink.zanxiangnet.ad.monitoring.kafka.KafkaComponent;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.flink.api.common.eventtime.*;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.DataStreamSource;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.windowing.assigners.TumblingEventTimeWindows;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.util.OutputTag;

import java.time.Duration;
import java.util.*;

@Slf4j
public class AdHourStatJob {

    /**
     * 可能有数据的最早日期
     */
    private static final String OLDEST_DAY = "2019-01-01";

    public static void main(String[] args) throws Exception {
        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();

        // 加载配置文件到 flink的全局配置中
        Properties props = new Properties();
        props.load(AdHourStatJob.class.getResourceAsStream("/application.properties"));
        Configuration configuration = new Configuration();
        props.stringPropertyNames().forEach(key -> {
            String value = props.getProperty(key);
            configuration.setString(key.trim(), StringUtils.isBlank(value) ? "" : value.trim());
        });
        env.getConfig().setGlobalJobParameters(configuration);

        KafkaSource<String> adStreamOfMinuteSource = KafkaComponent.buildKafkaSource(props, KafkaComponent.KafkaTopic.adHourTopic, KafkaComponent.KafkaTopic.KafkaGroupId.adHourConsumerGroup);
        DataStreamSource<String> adStreamOfMinuteIn = env.fromSource(adStreamOfMinuteSource, WatermarkStrategy.noWatermarks(), "adHourSource_kafka");

        // 广告分钟数据（前 5分钟的广告消耗数据）
        final OutputTag<AdDataOfMinuteODS> adMinuteStreamTag = new OutputTag<AdDataOfMinuteODS>("adMinuteStream") {
        };
        // 广告小时数据（往前回滚 10天）
        final OutputTag<AdDataOfHourODS> adHourStreamTag = new OutputTag<AdDataOfHourODS>("adHourStream") {
        };
        // 广告小时数据（从分钟数据流切出来的整点数据）
        final OutputTag<AdStatOfHourDWD> adHourFromMinuteStreamTag = new OutputTag<AdStatOfHourDWD>("adHourFromMinuteStream") {
        };

        // 对流进行映射，拆分（实时的分钟流和回滚的小时流）
        SingleOutputStreamOperator<AdDataOfMinuteODS> adODSStream = adStreamOfMinuteIn.filter(StringUtils::isNotBlank)
                .process(new AdHourDTOStreamProcess(adMinuteStreamTag, adHourStreamTag));

        // 分钟流
        DataStream<AdDataOfMinuteODS> adMinuteODSStream = adODSStream.getSideOutput(adMinuteStreamTag);
        // 分钟流-写入原始表
        new KeyedBatchStream<>("adMinuteODSStream", adMinuteODSStream.keyBy(AdDataOfMinuteODS::getStatDay), 4000L, 2 * 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdDataOfMinuteODS.class))
                .name("sink_ad_minute_ods");

        // 分钟流-计算
        SingleOutputStreamOperator<AdStatOfMinuteDWD> adMinuteDWDStream = adMinuteODSStream
                // 打水印，允许数据延迟 6分钟，同时指定时间流
                .assignTimestampsAndWatermarks(WatermarkStrategy.<AdDataOfMinuteODS>forBoundedOutOfOrderness(Duration.ofMinutes(6L))
                        .withTimestampAssigner((SerializableTimestampAssigner<AdDataOfMinuteODS>) (adODS, l) -> adODS.getStatTime()))
                .keyBy(AdDataOfMinuteODS::getAdId)
                // 开一个 5分钟的滚动窗口
                .window(TumblingEventTimeWindows.of(Time.minutes(5L)))
                .trigger(new AdMinuteODSStreamTrigger())
                .process(new AdMinuteDWDProcess(adHourFromMinuteStreamTag));
        new KeyedBatchStream<>("adMinuteDWDStream", adMinuteDWDStream.keyBy(AdStatOfMinuteDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfMinuteDWD.class))
                .name("sink_ad_minute_dwd");

        //cost----分钟数据处理
        SingleOutputStreamOperator<CostMinuterDM> clickhouseMinuteDmStream =
                adMinuteDWDStream
                        .keyBy(AdStatOfMinuteDWD::getAdId)
                        .window(TumblingEventTimeWindows.of(Time.minutes(5L)))
                        .trigger(new CostMinuteDMStreamTrigger())
                        .process(new CostMinuteProcess())
                        .name("sink_ad_minute_dm_clickhouse");

        BatchSinkMinute batchSinkMinute = new BatchSinkMinute();
        clickhouseMinuteDmStream.addSink(batchSinkMinute);


        //cost----小时数据处理
        /*SingleOutputStreamOperator<CostHourDM> clickhouseMinuteHourDmStream =
                adMinuteDWDStream
                        .keyBy(AdStatOfMinuteDWD::getAdId)
                        .window(TumblingEventTimeWindows.of(Time.minutes(5L)))
                        .trigger(new CostMinuteDMStreamTrigger())
                        .process(new CostMinuteHourProcess())
                        .name("sink_ad_minute_hour_dm_clickhouse");

        BatchSinkHour batchSinkMinuteHour = new BatchSinkHour();
        clickhouseMinuteHourDmStream.addSink(batchSinkMinuteHour);*/


        // 小时流
        DataStream<AdDataOfHourODS> adHourODSStream = adODSStream.getSideOutput(adHourStreamTag);
        // 小时流-写入原始表
        new KeyedBatchStream<>("adHourODSStream", adHourODSStream.keyBy(AdDataOfHourODS::getStatDay), 4000L, 5 * 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdDataOfHourODS.class))
                .name("sink_ad_hour_ods");

        // 小时流-计算
        SingleOutputStreamOperator<AdStatOfHourDWD> adHourDWDStream =
                adHourODSStream.keyBy(AdDataOfHourODS::getAdId)
                        .countWindow(1)
                        .process(new AdHourDWDProcess());
        DataStream<AdStatOfHourDWD> adHourDWDAllStream = adMinuteDWDStream.getSideOutput(adHourFromMinuteStreamTag).union(adHourDWDStream);
        new KeyedBatchStream<>("adHourDWDStream", adHourDWDAllStream.keyBy(AdStatOfHourDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfHourDWD.class))
                .name("sink_ad_hour_dwd");

        //小时数据
        SingleOutputStreamOperator<CostHourDM> clickhouseHourDmStream =
                adHourDWDAllStream
                        .keyBy(AdStatOfHourDWD::getAdId)
                        .countWindow(1)
                        .process(new CostHourProcess())
                        .name("sink_ad_hour_dm_clickhouse");

        BatchSinkHour batchSinkhour = new BatchSinkHour();
        clickhouseHourDmStream.addSink(batchSinkhour);


        env.execute("ad_hour_stream_job");
    }
}
