package flink.zanxiangnet.ad.monitoring.process;

import flink.zanxiangnet.ad.monitoring.config.ClickhouseDataSourceFactory;
import flink.zanxiangnet.ad.monitoring.dao.mapper.AdStatOfDayDWDMapper;
import flink.zanxiangnet.ad.monitoring.pojo.entity.*;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.util.Collector;
import org.apache.flink.util.OutputTag;
import org.apache.ibatis.datasource.DataSourceFactory;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;

@Slf4j
public class AdMinuteDWDProcess extends ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow> {
    private static final DateTimeFormatter formatForLastReduceKey = DateTimeFormatter.ofPattern("yyyy-MM-dd_HH:mm");

    private final OutputTag<AdStatOfHourDWD> adHourFromMinuteStreamTag;

    private SqlSessionFactory sqlSessionFactory;
    // 历史的天数据
    private ValueState<AdStatOfDayDWD> historyDayState;
    // 上次查询的天数据
    private ValueState<String> lastQueryDayState;
    // 之前聚合的昨天的数据
    private MapState<String, AdStatOfMinuteDWD> yesterdayMinuteState;
    // 前 5分钟聚合的数据
    private MapState<String, AdStatOfMinuteDWD> lastReduceState;

    public AdMinuteDWDProcess(OutputTag<AdStatOfHourDWD> adHourFromMinuteStreamTag) {
        this.adHourFromMinuteStreamTag = adHourFromMinuteStreamTag;
    }

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();

        Properties ckProps = new Properties();
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_URL, params.get(ApplicationProperties.CK_URL));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_USER, params.get(ApplicationProperties.CK_USERNAME));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_PASSWORD, params.get(ApplicationProperties.CK_PASSWORD));

        DataSourceFactory dataSourceFactory = new ClickhouseDataSourceFactory();
        dataSourceFactory.setProperties(ckProps);
        Environment environment = new Environment("clickhouse", new JdbcTransactionFactory(), dataSourceFactory.getDataSource());
        org.apache.ibatis.session.Configuration configuration = new org.apache.ibatis.session.Configuration(environment);
        // 开启驼峰规则
        configuration.setMapUnderscoreToCamelCase(true);
        configuration.getTypeAliasRegistry().registerAlias(AdStatOfDayDWD.class);
        // addMapper一定要放到 alias的后面！！！！！
        configuration.addMapper(AdStatOfDayDWDMapper.class);
        sqlSessionFactory = new SqlSessionFactoryBuilder().build(configuration);

        historyDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("historyDayState", AdStatOfDayDWD.class));
        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
        yesterdayMinuteState = getRuntimeContext().getMapState(new MapStateDescriptor<>("yesterdayMinuteState", String.class, AdStatOfMinuteDWD.class));
        lastReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("lastReduceState", String.class, AdStatOfMinuteDWD.class));
    }

    @Override
    public void process(Long elementCount, ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>.Context context,
                        Iterable<AdDataOfMinuteODS> iterable, Collector<AdStatOfMinuteDWD> collector) throws Exception {
        long beginTime = context.window().getStart();
        LocalDateTime beginDateTime = DateUtil.milliToLocalDateTime(beginTime);
        LocalDate beginDate = beginDateTime.toLocalDate();
        String statDay = DateUtil.formatLocalDate(beginDate);
        int hour = beginDateTime.getHour();
        long now = context.currentProcessingTime();
        AdDataOfMinuteODS statODS = null;
        List<AdDataOfMinuteODS> adDataOfMinuteODSList = new ArrayList<>(24);
        for (AdDataOfMinuteODS adDataOfMinuteODS : iterable) {
            adDataOfMinuteODSList.add(adDataOfMinuteODS);
            if (adDataOfMinuteODS.getHour() != hour) {
                continue;
            }
            statODS = adDataOfMinuteODS;
        }
        if (statODS == null) {
            return;
        }

        Long adId = statODS.getAdId();

        // 昨天聚合的数据
        AdStatOfMinuteDWD yesterdayMinuteDWD = yesterdayMinuteState.get(DateUtil.formatLocalDate(beginDate.minusDays(1L)));
        // 之前的数据
        String lastQueryDay = lastQueryDayState.value();
        if (lastQueryDay == null || !lastQueryDay.equals(statDay)) {
            try (SqlSession session = sqlSessionFactory.openSession()) {
                AdStatOfDayDWDMapper mapper = session.getMapper(AdStatOfDayDWDMapper.class);
                List<AdStatOfDayDWD> historyDayData = mapper.lastReduceResult(adId, null, DateUtil.formatLocalDate(beginDate.minusDays(2L)), 1);
                if (!historyDayData.isEmpty()) {
                    historyDayState.update(historyDayData.get(historyDayData.size() - 1));
                }
                lastQueryDayState.update(statDay);
            }
        }
        AdStatOfDayDWD beforeYesterdayDayDWD = historyDayState.value();

        AdStatOfMinuteDWD lastReduceData = lastReduceState.get(beginDateTime.minusMinutes(5L).format(formatForLastReduceKey));
        // 聚合当天的全部数据
        AdStatOfMinuteDWD newAdStat = AdStatOfMinuteDWD.reduce(beforeYesterdayDayDWD, yesterdayMinuteDWD, adDataOfMinuteODSList, statODS, lastReduceData, now);
        collector.collect(newAdStat);
        if (beginDateTime.getMinute() == 55) {
            context.output(adHourFromMinuteStreamTag, AdStatOfHourDWD.byMinuteDWD(newAdStat));
        }
        lastReduceState.put(beginDateTime.format(formatForLastReduceKey), newAdStat);
        yesterdayMinuteState.put(statDay, newAdStat);

        List<String> removeKeys = new ArrayList<>(10);
        Iterator<Map.Entry<String, AdStatOfMinuteDWD>> lastIterator = lastReduceState.iterator();
        while (lastIterator.hasNext()) {
            Map.Entry<String, AdStatOfMinuteDWD> temp = lastIterator.next();
            if (temp.getKey().equals(beginDateTime.format(formatForLastReduceKey))
                    || temp.getKey().equals(beginDateTime.minusMinutes(5L).format(formatForLastReduceKey))
                    || temp.getKey().equals(beginDateTime.minusMinutes(10L).format(formatForLastReduceKey))) {
                continue;
            }
            removeKeys.add(temp.getKey());
        }
        if (!removeKeys.isEmpty()) {
            for (String key : removeKeys) {
                lastReduceState.remove(key);
            }
        }
        removeKeys.clear();
        Iterator<Map.Entry<String, AdStatOfMinuteDWD>> yesterdayIterator = yesterdayMinuteState.iterator();
        while (yesterdayIterator.hasNext()) {
            Map.Entry<String, AdStatOfMinuteDWD> temp = yesterdayIterator.next();
            if (temp.getKey().equals(DateUtil.formatLocalDate(beginDate))
                    || temp.getKey().equals(DateUtil.formatLocalDate(beginDate.minusDays(1L)))
                    || temp.getKey().equals(DateUtil.formatLocalDate(beginDate.minusDays(2L)))) {
                continue;
            }
            removeKeys.add(temp.getKey());
        }
        if (!removeKeys.isEmpty()) {
            for (String key : removeKeys) {
                yesterdayMinuteState.remove(key);
            }
        }
    }

    @Override
    public void clear(ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>.Context context) throws Exception {
    }
}
